\ SD Driver

\ SD-SPI Variables initialize by the boot code
: spi_var create , does> @ spi_vars + ;
 0 spi_var spi_engine_cog
 4 spi_var spi_command
 8 spi_var spi_block_index
12 spi_var spi_buffer_address
16 spi_var spi_rootdir
20 spi_var spi_filesystem
24 spi_var spi_clustershift
28 spi_var spi_dataregion

\ Read a single 512-byte sector
: spi_readblock ( block_index buffer_address -- retval )
  spi_engine_cog @ 0= if 2drop -999 exit then   \ ERR_SPI_ENGINE_NOT_RUNNING
  dup 3 and if 2drop -4 exit then               \ ERR_BLOCK_NOT_LONG_ALIGNED
  spi_buffer_address !
  spi_block_index !
  [char] r spi_command !
  begin spi_command @ [char] r <> until         \ Wait for command to complete
  spi_command @
;
  
\ Write a single 512-byte sector
: spi_writeblock ( block_index buffer_address -- retval )
  spi_engine_cog @ 0= if 2drop -999 exit then   \ ERR_SPI_ENGINE_NOT_RUNNING
  dup 3 and if 2drop -4 exit then               \ ERR_BLOCK_NOT_LONG_ALIGNED
  spi_buffer_address !
  spi_block_index !
  [char] w spi_command !
  begin spi_command @ [char] w <> until         \ Wait for command to complete
  spi_command @
;

\ 512-byte sector buffer
create spi_buf 512 allot

\ File information structure
create file_info 5 cells allot
: file_var create , does> @ file_info + ;
 0 file_var file_sector0
 4 file_var file_length
 8 file_var file_position
12 file_var file_sector
16 file_var file_remaining

\ Get an unaligned word
: brword dup 1+ swap c@ swap c@ 8 lshift or ;

\ Get an unaligned long
: brlong dup brword swap 2 + brword 16 lshift or ;

\ List the files in the root directory
: list-files
  spi_rootdir @
  begin
    dup spi_buf spi_readblock drop
    512 0 do
      i spi_buf + dup c@
      if
        dup 11 type bl emit
        \ dup 11 + c@ .
        \ dup 26 + brword .
        dup 28 + brlong .
        cr drop \ 11 + 21 dump
      else
        r> r> 2drop 2drop exit
      then
    32 +loop
    1+
  again
;

\ Convert a character to upper case
: toupper dup [char] a >=
  if
    dup [char] z <= 
    if
      32 -
    then
  then
;

create fatname 12 allot

\ Convert an 8.3 file name to FAT format
: convertfname ( addr num -- )
  fatname 11 bl fill
  8 0 do
    dup 1 <
    if 2drop r> r> 2drop exit then
    over 1+ rot c@ dup [char] . =
    if drop swap 1- leave
    else
      toupper i fatname + c!
      swap 1-
    then
  loop
  over c@ [char] . =
  if 1- swap 1+ swap then
  11 8 do
    dup 1 <
    if 2drop r> r> 2drop exit then
    over 1+ rot c@
    toupper i fatname + c!
    swap 1-
  loop
  2drop
;

\ Compare two strings and return 0 if they match
: compare ( str1 len1 str2 len2 ... n )
  rot over <> if 2drop drop 1 exit then
  0 ?do
    2dup c@ swap c@ <>
    if 2drop r> r> 2drop 1 exit then
    1+ swap 1+
  loop
  2drop 0
;

\ Attempt to open a file and return 0 on success
: open-file ( addr len fam -- fileid ior )
  drop convertfname
  spi_rootdir @
  begin
    dup spi_buf spi_readblock drop
    512 0 do
      i spi_buf + dup c@
      if
        dup 11 fatname 11 compare 0=
        if
          \ dup 11 type bl emit
          \ dup 11 + c@ .
          \ dup 26 + brword .
          \ dup 28 + brlong .
          \ dup 26 + brword spi_clustershift @ lshift dup .
          \ spi_dataregion @ + .
          \ cr
          dup 26 + brword spi_clustershift @ lshift spi_dataregion @ +
          swap 28 + brlong swap
          r> r> 2drop rot drop
          file_sector0 !
          file_length !
          0 file_position !
          file_sector0 @ file_sector !
          file_length @ 512 min file_remaining !
          file_sector @ spi_buf spi_readblock drop
          1 0 exit
        then
        drop
      else
        r> r> 2drop 2drop 1 1 exit
      then
    32 +loop
    1+
  again
;

\ Read one byte from the file.  Return -1 on EOF
: file-getc
  file_position @ file_length @ >= if -1 exit then
  file_remaining @ 1 <
  if
    file_sector @ 1+ file_sector !
    file_sector @ spi_buf spi_readblock drop
    file_length @ file_position @ - 512 min file_remaining !
  then
  file_position @ 511 and spi_buf + c@
  file_position @ 1+ file_position !
  file_remaining @ 1- file_remaining !
;

\ Read one line from the file. Return number of bytes read or -1 on EOF
: read-line ( addr size fileid -- len flag ior )
  drop >r dup
  begin
    \ ." Trace 1" cr
    r@ 1 < if r> drop swap - -1 0 exit then
    file-getc dup -1 =
    if
      \ ." Trace 2" cr
      r> 2drop swap - dup
      if -1 0 exit then
      \ ." Trace 3" cr
      0 0 exit
    else
      \ ." Trace 4" cr
      dup dup 10 = swap 13 = or
      if r> 2drop swap - -1 0 exit then
      \ ." Trace 5" cr
      over c! 1+
      r> 1- >r
    then
  again
;

\ Type the file to the console
: type-file bl word count 0 open-file swap drop
  if
    ." Could not open file" cr
  else
    begin
      file-getc dup -1 = if drop exit then
      dup 10 =
      if
        drop cr
      else
        emit
      then
    again
  then
;

\ Line Type the file to the console
: file-typel bl word count 0 open-file swap drop
  if
    ." Could not open file" cr
  else
    begin
      pad 100 0 read-line drop
      '.s cr
      0 = if drop exit then
      pad swap type cr
    again
  then
;

\ Create an new definition for refill
: _refill1 tib 200 source-id
  if
    0 read-line drop
    0 = if drop 0 to source-id 0 then
  else
    accept
  then
  #tib ! 0 >in !
;

' _refill1 is refill

\ Include file
: include bl word count 0 open-file
  if
    drop ." Could not open file" cr
  else
    to source-id
  then
;

\ Get the file size
: file-size ( fileid -- size ior ) drop file_length @ 0 ; 

\ Get the file postion
: file-position ( fileid -- position ior ) drop file_position @ 0 ;

\ Repostion the file
: reposition-file ( position fileid -- ior ) drop
  dup 0 < if 1 exit then
  dup file_length @ >= if 1 exit then
  dup -512 and file_position @ -512 and <>
  if
    dup 9 rshift file_sector0 @ +
    dup file_sector !
    spi_buf spi_readblock drop
  then
  dup file_position !
  file_length @ swap - 512 min file_remaining !
  0
;
